/*
 *  Sort, or -- depending on compilation options -- shuffle the
 *  lines in a text file.  (Yes, two programs which are inverses
 *  of each other share the same main.c file.)
 *
 *  This is an implementation of the classic external sort,
 *  sorting as much data as will fit in memory, but if necessary
 *  (i.e. if the complete input won't fit in memory), writing sorted
 *  chunks to temporary files and then merging those together.
 *
 *  This file is the driver: parsing the command line and
 *  performing the very highest level of the external sort algorithm.
 *  But the real work is all elsewhere: reading lines, sorting
 *  in-memory chunks, merging temporary files, and writing output
 *  are all done fairly modularly, via subsidiary functions.
 *  See also input.c.
 *
 *  By using variants on those subsidiary functions, it is
 *  possible to (a) sort db records or data objects other than lines,
 *  and (b) shuffle rather than sort.
 *
 *  One wrinkle of the external sort implementation deserves mention.
 *  Rather than trying to predict in advance how much memory will be
 *  available and therefore how big each in-memory chunk can be,
 *  the program uses a "bop 'til you drop" strategy: keep reading
 *  lines and copying them into malloc'ed memory, until malloc fails,
 *  at which point you know that how much memory you have is enough
 *  for just one less than the number of lines you've read.
 *  (One of Dave Barry's handy cooking tips for men pops to mind:
 *  "Cook in a covered pan in a medium oven until just before
 *  copious quantities of black smoke fill the kitchen.")
 *
 *  Though this strategy nicely dodges the nuisancy and nonportable
 *  problem of trying to quantify the available memory in advance
 *  (a problem which has no decent or portable solution), there
 *  remain two difficulties.  One is simply that the determination
 *  to persevere (as opposed to printing "out of memory" and exiting)
 *  in the case of a malloc failure is tricky.  The second is that
 *  in a virtual memory environment, using all seemingly available
 *  memory can lead to surprisingly poor performance.
 *
 *  To resolve the second problem, the code can be compiled to
 *  constrain itself to some predefined limit.  If the preprocessor
 *  macro VMLIMIT is defined (as it is by default in the distributed
 *  Makefile), the code will not, in fact, keep doggedly allocating
 *  memory until malloc fails, but will instead voluntarily go to
 *  a temporary file (as if malloc had failed) when the number of
 *  bytes so far allocated reaches the value of the VMLIMIT macro.
 *  (See aditional comments in lmalloc.c.)
 *
 *  Notice that this is *not* a capitulation against the goal of
 *  being portable and of not trying to predetermine the amount of
 *  available memory.  If we were trying to predetermine the amount
 *  of available memory, we'd get poor performance if we picked that
 *  number much too low, and probably failures if we picked that number
 *  too big, and in any case picking that number would be a huge nuisance.
 *  What VMLIMIT is, on the other hand, is an amply large number which
 *  might in fact be bigger than the amount of memory actually,
 *  i.e. malloc might fail before VMLIMIT is reached, but that's okay.
 *  (In fact, the implementation of VMLIMIT works by simulating a
 *  malloc failure, which the rest of the code is of necessity already
 *  all set up to deal with.)  Picking a wrong value of VMLIMIT will
 *  never lead to failures, only to possibly poor performance if it's
 *  too small or too large.
 *
 *  I don't honestly know what a good value of VMLIMIT is these days.
 *  When I first wrote the VMLIMIT code, it was because a large
 *  attempted sort had run for an hour and a half before crashing the
 *  machine it was running on (which was my ISP's timesharing shell
 *  server; rather an embarrassing mistake).  Setting a limit of two
 *  megabytes not only avoided the crash but allowed the sort to
 *  finish in minutes.  (Evidently the special-purpose external sort
 *  algorithm can be *much* more efficient than general-purpose virtual
 *  memory algorithms when faced with the thrashing imposed by
 *  a too-large in-memory sort.)  On a modern machine, especially if
 *  single-user, you could obviously afford a much higher limit than
 *  2 meg before swapping at all, let alone thrashing.  So feel free
 *  to up the value of VMLIMIT, although if you find (as I do) that
 *  performance is adequate with it as it is, you can just leave it.
 *
 *  Copyright 1988-2003
 *  Steve Summit, scs@eskimo.com
 *  This code may be freely redistributed and used
 *  so long as the author's name and this notice remain intact,
 *  and any modifications are marked as such.
 *
 *  See http://www.eskimo.com/~scs/src for full documentation
 *  and possible updates.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#ifdef SHUFFLE
#include <time.h>
#endif
#include <errno.h>

#ifdef mac
#include <console.h>
#endif

#ifdef DBSORT
#include <db.h>
#endif

#include "defs.h"
#include "sort.h"
#include "sortmain.h"
#include "tmpfile.h"
#include "string2.h"

#ifdef SHUFFLE
unsigned int seed;
int haveseed = FALSE;
char *seedfile = NULL;
#endif

#ifdef SHUFFLE
char *progname = "shuffle";
#else
#ifdef DBSORT
char *progname = "dbsort";
#else
char *progname = "sort";
#endif
#endif

char usage[] = "usage: %s %s [-o outputfile] [inputfiles ...]\n";

#ifdef SHUFFLE
char options[] = "[options]";
#else
#ifdef DBSORT
char options[] = "[-dfFnru] -k key";
#else
char options[] = "[-dfFnru] [+col1 +col2 ...]";
#endif
#endif

struct column *columns = NULL;

int uflag = FALSE;
int defltflags = 0;

int keepgoing = FALSE;
int onlyselected = FALSE;

int tabchar = -1;	/* means whitespace */

int usedtemps = FALSE;

#ifdef DBSORT

#define MAXSORTKEYS 5

char *sortkeys[MAXSORTKEYS];
int nsortkeys = 0;

int dbfflags = DB_BLANKSEP;
int keyfoldcase = FALSE;
int preserveindent = FALSE;
int dbcontin = FALSE;
int dupkeyok = FALSE;
int mailflag = FALSE;
char *mailsepkey = "From ";

dbdesc dbd;

#endif

#ifdef ESUITE
int maxchunk = 0;
int maxtmpfd = 0;
#endif

extern char *agetline();

struct tmpfile *tmpbase = NULL;
struct tmpfile *nextemp = NULL;

#ifdef SHUFFLE
void saveseed();
#endif

extern int compare();

#ifdef DBSORT
extern int dbcompare();
#endif

#ifndef unix
#ifdef GETPROGNAME
extern char *getprogname();
#endif
#endif

main(argc, argv)
int argc;
char *argv[];
{
int argi;
int didone = FALSE;
#ifndef DBSORT
FILE *fd;
#else
struct dbfd *fd;
int usedstdin = FALSE;
#endif
FILE *ofd;
char *p;
#ifndef SHUFFLE
struct column **endp = &columns;
#endif
struct column *cp = NULL;
char *outputfile = NULL;
int printedhelp = FALSE;
int firsttime = TRUE;

#ifdef mac
argc = ccommand(&argv);
#endif

#ifdef unix

if(argc > 0)
	{
	progname = argv[0];
	if((p = strrchr(progname, '/')) != NULL)
		progname = p + 1;
	}

#else
#ifdef GETPROGNAME

progname = getprogname(argc, argv, progname);

#endif
#endif

stashalloc();	/* stash some memory away now,
			which we'll need (to recover) if we run out */

tmpbase = nextemp = tfalloc();

#ifdef DBSORT
dbd = db_alloc(0);
#endif

for(argi = 1; argi < argc; argi++)
	{
	if((argv[argi][0] == '-' || argv[argi][0] == '+') &&
						argv[argi][1] != '\0')
		{
		int *flagp;

		cp = NULL;

#ifndef SHUFFLE

		if(argv[argi][0] == '+' && isdigit(argv[argi][1]))
			{
			int n, n2;

			n = strtol(&argv[argi][1], &p, 10);
			if(*p == '-')
				n2 = strtol(p + 1, &p, 10);
			else	n2 = n;

			cp = (struct column *)malloc(sizeof(struct column));
			if(cp == NULL)
				{
				fprintf(stderr, "%s: out of memory\n",
								progname);
				exit(1);
				}

			cp->c_number = n;
			cp->c_n2 = n2;				/* XXX */
			cp->c_flags = defltflags; /* should it get later ones? */
			flagp = &cp->c_flags;
			cp->c_next = NULL;
			*endp = cp;
			endp = &cp->c_next;
			}
		else
#endif
			{
			p = &argv[argi][1];
			flagp = &defltflags;
			}

		for(; *p != '\0'; p++)
			{
			switch(*p)
				{
				case '?':
printhelp:	fprintf(stderr, usage, progname, options);
#ifndef SHUFFLE
		fprintf(stderr, "+#[-#]\tsort column # (0-based; several may appear; per-column flags may follow)\n");
		fprintf(stderr, "options:\n");
#ifdef notyet
		fprintf(stderr, "-b\tskip leading blanks\n");
		fprintf(stderr, "-c\tcheck input already sorted\n");
		fprintf(stderr, "-d\tdictionary sort order\n");
#endif
#ifdef DBSORT
		fprintf(stderr, "-contin\tallow continuation lines\n");
#endif
#ifdef SORTDATETIME
		fprintf(stderr, "-d\tdate/time sort\n");
#endif
#ifdef DBSORT
		fprintf(stderr, "-dkok\tduplicate keys OK\n");
#endif
		fprintf(stderr, "-f\tfold upper/lower case (i.e. caseless)\n");
		fprintf(stderr, "-F\t\"filename\" sort\n");
#endif
		fprintf(stderr, "-h, -?\tprint this help message\n");
#ifdef DBSORT
		fprintf(stderr, "-hc\tallow continuation lines\n");
#endif
#ifndef SHUFFLE
#ifdef notyet
		fprintf(stderr, "-i\tignore non-ASCII characters\n");
#endif
#ifdef DBSORT
		fprintf(stderr, "-k key\tset field to sort on\n");
		fprintf(stderr, "-ki\tignore case in field keywords\n");
#endif
#endif
#ifdef notyet
		fprintf(stderr, "-m\tmerge already-sorted inputs\n");
#endif
#ifdef DBSORT
		fprintf(stderr, "-mail\tread mailbox format\n");
#ifdef notyet
		fprintf(stderr, "-mailh key\tmailbox format, separator key\n");
#endif
#endif
#ifdef ESUITE
		fprintf(stderr, "-maxchunk n  max in-memory sort lines n\n");
		fprintf(stderr, "-maxtmpfd n  max tmp file descriptors n\n");
#endif
#ifndef SHUFFLE
		fprintf(stderr, "-n\tnumeric sort\n");
#endif
		fprintf(stderr, "-o file\toutput to file\n");
#ifdef DBSORT
		fprintf(stderr, "-pi\tpreserve indentation\n");
#endif
#ifndef SHUFFLE
		fprintf(stderr, "-r\treverse sort\n");
#ifdef notyet
		fprintf(stderr, "-s\t\"stable\" sort (preserve original ordering for duplicates)\n");
		fprintf(stderr, "-tc\tset tab character to c\n");
#endif
#endif
#ifdef SHUFFLE
		fprintf(stderr, "-seed s\tuse initial seed s\n");
		fprintf(stderr, "-seedfile f  keep seed state in f\n");
#endif
#ifdef notyet
		fprintf(stderr, "-T dir\tset tmp directory\n");
#endif
#ifndef SHUFFLE
		fprintf(stderr, "-u\tunique output (suppress duplicates)\n");
#endif
					printedhelp = TRUE;
					break;
#ifndef SHUFFLE
#ifdef notyet
				case 'b':
					*flagp |= SKIPBLANKS;
					break;
#endif

				case 'c':
#ifdef DBSORT
					if(*(p+1) == 's')
						{
						p++;
						dbfflags |= DB_KEYCHAR;
						break;
						}
					else if(strcmp(p, "contin") == 0)
						{
						dbcontin = TRUE;
						p = "x";    /* short circuit */
						break;
						}
#endif
#ifdef notyet
					checkflag = TRUE;
#else
					goto badopt;
#endif
					break;

				case 'd':
#ifdef DBSORT
					if(stricmp(p, "dkok") == 0)
						{
						dupkeyok = TRUE;
						p = "x";    /* short circuit */
						break;
						}
#endif
#ifdef notyet
					*flagp |= DICTIONARY;
#endif
#ifdef SORTDATETIME
					*flagp = (*flagp & ~TYPEMASK) | DATETIME;
#endif
					break;

				case 'f':
					*flagp |= FOLD;
					break;

				case 'F':	/* poor choice */
					*flagp = (*flagp & ~TYPEMASK) | FILENAME;
					break;
#endif /* !SHUFFLE */
				case 'h':
#ifdef DBSORT
					if(*(p+1) == 'c')
						{
						dbcontin = TRUE;
						p++;
						break;
						}
#endif
				/* FALLTHROUGH */
				case 'H':
					if(stricmp(p, "help") == 0)
						p = "x";    /* short circuit */
					goto printhelp;
#ifndef SHUFFLE
#ifdef notyet
				case 'i':
					*flagp |= IGNORE;
					break;
#endif
#ifdef DBSORT
				case 'k':
					if(strcmp(p, "ki") == 0)
						{
						keyfoldcase = TRUE;
						p = "x";    /* short circuit */
						break;
						}

					if(nsortkeys >= MAXSORTKEYS)
						{
						fprintf(stderr,
							"%s: too many sort keys\n",
								progname);
						break;
						}

					sortkeys[nsortkeys++] = argv[++argi];
					break;
#endif
#endif
				case 'm':
#ifdef DBSORT
					if(strcmp(p, "mail") == 0)
						{
						mailflag = TRUE;
						p = "x";    /* short circuit */
						}
#ifdef notyet
					else if(strcmp(p, "mailh") == 0)
						{
						mailflag = TRUE;
						mailsepkey = argv[++argi];
						p = "x";    /* short circuit */
						}
#endif
					else
#endif
#ifdef ESUITE
					     if(strcmp(p, "maxchunk") == 0)
						{
						maxchunk = atoi(argv[++argi]);
						p = "x";    /* short circuit */
						}
					else if(strcmp(p, "maxtmpfd") == 0)
						{
						maxtmpfd = atoi(argv[++argi]);
						p = "x";    /* short circuit */
						}
					else
#endif
						{
#ifdef notyet
						mergeflag = TRUE;
#else
						goto badopt;
#endif
						}
					break;
#ifndef SHUFFLE
				case 'n':
					*flagp = (*flagp & ~TYPEMASK) | NUMERIC;
					break;
#endif
				case 'o':
					outputfile = argv[++argi];
					break;
#ifdef DBSORT
				case 'p':
					if(strcmp(p, "pi") == 0)
						{
						preserveindent = TRUE;
						p = "x";    /* short circuit */
						break;
						}
					else	goto badopt;
#endif
#ifndef SHUFFLE
				case 'r':
					*flagp |= REVERSE;
					break;
#endif
				case 's':
#ifdef SHUFFLE
					if(strcmp(p, "seed") == 0)
						{
#ifdef pdp11
						seed = atoi(argv[++argi]); /* XXX */
#else
						seed = (unsigned int)
							strtoul(argv[++argi],
							    (char **)NULL, 10);
#endif
						haveseed = TRUE;
						p = "x";
						break;
						}
					else if(strcmp(p, "seedfile") == 0)
						{
						seedfile = argv[++argi];
						p = "x";
						break;
						}
#endif
#ifndef SHUFFLE
#ifdef notyet
					stable = TRUE;
					break;
#else
					goto badopt;
#endif
#else
					goto badopt;
#endif

#ifndef SHUFFLE
				case 't':
					if(*(p+1) != '\0')
						tabchar = *++p;
					else	tabchar = argv[++argi][0];	/* XXX what if it's a string? */
					break;
#endif
#ifdef notyet
				case 'T':
					set tmp directory
					break;
#endif
#ifndef SHUFFLE
				case 'u':
					uflag = TRUE;
					onlyselected = TRUE;
					break;
#endif
#ifdef DBSORT
				case 'w':
					if(*(p+1) == 's')
						{
						p++;
						dbfflags |= DB_KEYBLANK;
						}
					else	goto badopt;
					break;
#endif
				default:
badopt:					fprintf(stderr,
						"%s: unknown option -%c\n",
								progname, *p);
				}
			}

		continue;
		}

#ifdef notdef

	for(cp = columns; cp != NULL; cp = cp->c_next)		/* ??? */
		cp->c_flags |= defltflags;			/* ??? */

#endif

	initialize();

	didone = TRUE;

#ifndef DBSORT

	if(strcmp(argv[argi], "-") == 0)
		fd = stdin;
	else	fd = fopen(argv[argi], "r");

#else

	usedstdin = FALSE;

	if(strcmp(argv[argi], "-") == 0)
		{
		fd = db_fdopen(stdin);
		usedstdin = TRUE;
		}
	else	fd = db_fopen(argv[argi], "r");

#endif

	if(fd == NULL)
		{
		fprintf(stderr, "%s: can't open %s: %s\n", progname,
						argv[argi], strerror(errno));
		exit(1);
		}

#ifdef DBSORT
	dbset2up(fd);
#endif

	input(fd, argv[argi]);

#ifndef DBSORT
	if(fd != stdin)
		(void)fclose(fd);
#else
	if(usedstdin)
		db_fdclose(fd);
	else	db_fclose(fd);
#endif
	}

if(!didone)
	{
	if(printedhelp)
		exit(0);

	initialize();

#ifndef DBSORT
	input(stdin, "standard input");
#else
	fd = db_fdopen(stdin),
	dbset2up(fd);
	input(fd, "standard input");
#endif
	}

stashfree();

if(usedtemps)
	anothertemp();		/* stash last chunk */
else	dosort();

if(outputfile == NULL)
	ofd = stdout;
else	{
	if((ofd = fopen(outputfile, "w")) == NULL)
		{
		fprintf(stderr, "%s: can't open %s: %s\n", progname,
						outputfile, strerror(errno));
		exit(1);
		}
	}

if(usedtemps)
	merge(ofd, outputfile);
else	output(ofd, outputfile);

if(ofd != stdout)
	{
	if(fclose(ofd) == EOF)
		{
		fprintf(stderr, "%s: error closing (writing?) %s: %s\n",
					progname, outputfile, strerror(errno));
		exit(1);
		}
	}

#ifdef SHUFFLE
saveseed();
#endif

return 0;
}

initialize()
{
static int beenhere = FALSE;

if(!beenhere)
	{
#ifdef DBSORT
	dbsetup();
#endif
#ifdef SHUFFLE
	randomize();
#endif
	beenhere = TRUE;
	}
}

#ifdef DBSORT

dbsetup()
{
if(nsortkeys == 0)
	{
	fprintf(stderr, "%s: must set sort key with -k\n", progname);
	exit(1);
	}

if(keyfoldcase)
	db_setopt(dbd, DB_CASELESS);

if(dupkeyok)
	db_setopt(dbd, DB_DUPKEYOK);

if(mailflag)
	{
	dbfflags = DB_KEYCHAR | DB_BLANKSEP | DB_XKEYSEP | DB_XKEYSEP |
							DB_RFC822 | DB_DEFTEXT;
	}

if(preserveindent)
	dbfflags |= DB_KEYNOSTRIP;

if(dbcontin)
	dbfflags |= DB_CONTIN;
}

dbset2up(dbfd)
struct dbfd *dbfd;
{
db_fsetdbf(dbfd, dbfflags);

if(mailflag)
	{
	db_fsetsepk(dbfd, mailsepkey);
	db_fschars(dbfd, ':', EOF);
	}
}

#endif

#ifndef DBSORT
extern char **array;
#else
extern dbnode *array;
#endif

extern int nalloc;
extern int nlines;

dosort()
{
#ifndef DBSORT
int elsize = sizeof(char *);
#else
int elsize = sizeof(dbnode);
#endif
#ifndef SHUFFLE
#ifndef DBSORT
int (*cmpfunc)() = compare;
#else
int (*cmpfunc)() = dbcompare;
#endif
#endif

#ifdef SHUFFLE
shuffle((char *)array, nlines, elsize);
#else
if(uflag)
	sort((char *)array, &nlines, elsize, cmpfunc, SORT_UNIQUE);
else	qsort((char *)array, nlines, elsize, cmpfunc);
#endif
}

output(fd, name)
FILE *fd;
char *name;
{
register int i;
#ifdef DBSORT
struct dbfd *dbofd = db_fdopen(fd);
db_fsetdbf(dbofd, dbfflags);
#endif

for(i = 0; i < nlines; i++)
	{
#ifndef DBSORT
	fprintf(fd, "%s\n", array[i]);
#else
	db_putent(dbofd, array[i]);
#endif

	if(ferror(fd))		/* can't trust fprintf return value */
		{
		if(name == NULL)
			fprintf(stderr, "%s: output write error: %s\n",
					progname, strerror(errno));
		else	fprintf(stderr, "%s: error writing %s: %s\n",
					progname, name, strerror(errno));
		exit(1);
		}
	}
}

static char *stash = NULL;

#define STASHSIZE 1024

/*
 *  Stash some memory away, so that if/when we run out completely,
 *  we can make it available for things like stdio output buffers,
 *  which we'll need while writing intermediate temporary files.
 */

stashalloc()
{
if(stash == NULL)
	stash = malloc(STASHSIZE);
}

stashfree()
{
if(stash != NULL)
	{
	free(stash);
	stash = NULL;
	}
}

#ifdef SHUFFLE

randomize()
{
int r;
#ifdef DEVRANDOM
int fd;
#endif

if(haveseed)
	{
	srand(seed);
	return;
	}

if(seedfile != NULL)
	{
	FILE *fp = fopen(seedfile, "r");
	if(fp != NULL)
		{
		r = fscanf(fp, "%u", &seed);
		fclose(fp);
		if(r == 1)
			{
			srand(seed);
			return;
			}
		}
	}

#ifdef DEVRANDOM

if((fd = open("/dev/urandom", 0)) >= 0)
	{
	r = read(fd, &seed, sizeof(seed));
	close(fd);
	if(r == sizeof(seed))
		{
		srand(seed);
		return;
		}
	}

#endif

seed = (unsigned int)time((time_t *)NULL);
#ifdef unix
seed ^= getpid() << 8;
#endif

srand(seed);
}

unsigned int fetchseed();

void
saveseed()
{
if(seedfile != NULL)
	{
	FILE *fp = fopen(seedfile, "w");
	if(fp == NULL)
		{
		fprintf(stderr, "%s: can't write seed state file %s: %s\n",
					progname, seedfile, strerror(errno));
		return;
		}

	fprintf(fp, "%u\n", fetchseed());

	fclose(fp);
	}
}

unsigned int
fetchseed()
{
return rand();		/* XXX imperfect, but portable and repeatable */
}

#endif
