/*
 *  Merge routines for external sort and shuffle programs.
 *  Open a bunch of temporary files, read first line from each,
 *  select first in-order line from among those read, output it,
 *  read another line from that file, lather, rinse, repeat.
 *
 *  Just as memory was a limited resource during the first,
 *  in-memory-sort phase of the operation, now file descriptors are
 *  potentially a limiting resource during merging.  If we had a
 *  truly monstrous input file to sort, such that we now have many,
 *  many partial temporary files, we may not be able to obtain enough
 *  file descriptors to keep them all open at once while merging them.
 *  In that case, we'll have to do partial merges to yet more temporary
 *  files, then merge those.
 *
 *  (Here, again, the determination of the number of available file
 *  descriptors is done empirically: when we can't open any more,
 *  and assuming the error is EMFILE or ENFILE, we've run out,
 *  and it's time to fall back to plan C.)
 *
 *  There's one remaining potential failure mode which is, as yet,
 *  not handled: if we have some vastly large number of temporary files,
 *  it's theoretically possible that we won't have enough memory to read
 *  the first line from all of them.  In practice, this seems unlikely
 *  to happen, since however limited memory is or isn't, file descriptors
 *  are likely to be more limited, meaning that we're likely to run out
 *  of file descriptors (if anything) first, and fall back to a multi-phase
 *  merge on a relatively small number of streams which we do in fact have
 *  enough memory to read one line from each of.  But you'll know you've
 *  hit this case if you ever get the message "aarrgh: case I can't handle:
 *  out of memory during merge".
 *  (It probably wouldn't be too hard to fix, at least in the case
 *  where the failure occurs while reading the first line, although
 *  handling it when reading a later line would be trickier.)
 *
 *  Copyright 1988-2003
 *  Steve Summit, scs@eskimo.com
 *  This code may be freely redistributed and used
 *  so long as the author's name and this notice remain intact,
 *  and any modifications are marked as such.
 *
 *  See http://www.eskimo.com/~scs/src for possible updates.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>

#ifdef DBSORT
#include <db.h>
#else
#include "agetline.h"
#endif

#include "defs.h"
#include "sortmain.h"
#include "tmpfile.h"

#ifdef DEBUG
static int nstage = 1;
#endif

extern struct tmpfile *tmpbase;

#ifdef ESUITE
extern int maxtmpfd;
static int ntmpfd = 0;
#endif

extern char *progname;

#ifdef DBSORT
extern dbdesc dbd;
extern struct dbnode *db_fgetent(struct dbfd *, struct db *, struct dbnode *, int, int *);	/* ? */
#endif

#ifndef SHUFFLE

#ifdef __STDC__
static insert(struct tmpfile *, struct tmpfile **);
#endif

static insert();

#endif

#ifdef SHUFFLE
/* XXX these shouldn't be global */
static int ntmp = 0;
static int nte = 0;
#endif

merge(ofd, name)
FILE *ofd;
char *name;
{
struct tmpfile *usetmp = tmpbase;
struct tmpfile *nxttmps = NULL;
struct tmpfile *tfp;
struct tmpfile *nexttfp;
#ifndef SHUFFLE
struct tmpfile *mergelist = NULL;
#endif
int usedmoretemps;
int r;

fpstash();

/* first, open & read one line from each tmp file, and build initial list */

again:	  /* come back here if we have to do n-stage merge */

usedmoretemps = FALSE;

for(tfp = usetmp; tfp != NULL; tfp = nexttfp)
	{
	nexttfp = tfp->tf_next;		/* grab early, 'cuz might free tfp */

	if(!(tfp->tf_flags & TF_INUSE))
		{
#ifndef SHUFFLE
		free((char *)tfp);
#endif
		continue;
		}

	errno = 0;

#ifndef DBSORT
	tfp->tf_fd = fopen(tfp->tf_filename, "r");
#else
	tfp->tf_dbfd = db_fopen(tfp->tf_filename, "r");
#endif

#ifdef ESUITE
	if(maxtmpfd != 0 && ++ntmpfd > maxtmpfd)
		{
#ifndef DBSORT
		fclose(tfp->tf_fd);
		ntmpfd--;
		tfp->tf_fd = NULL;
		errno = EMFILE;
#endif
		}
#endif

#ifndef DBSORT
	if(tfp->tf_fd == NULL)
#else
	if(tfp->tf_dbfd == NULL)
#endif
		{
		if(errno == 0 || errno == EMFILE || errno == ENFILE)
			{

#ifndef SHUFFLE
#ifdef DEBUG
			fprintf(stderr, "Yow! multi-stage merge (%d)\n", nstage);
#endif
			yetanothertemp(mergelist, &nxttmps);
			/* XXX gotta free old mergelist, etc. */
			usedmoretemps = TRUE;
			mergelist = NULL;
#ifndef DBSORT
			tfp->tf_fd = fopen(tfp->tf_filename, "r");
#else
			tfp->tf_dbfd = db_fopen(tfp->tf_filename, "r");
#endif
#ifdef ESUITE
			ntmpfd++;
#endif
#else
			/* I haven't figured out how to do multi-stage shuffle merge yet */
			fprintf(stderr, "%s: trying to merge too many temp files\n", progname);
			fprintf(stderr, "(can't handle yet)\n");
			exit(1);
#endif
			}
		}

#ifndef DBSORT
	if(tfp->tf_fd == NULL)
#else
	if(tfp->tf_dbfd == NULL)
#endif
		{
			/* XXX reindent */
			fprintf(stderr, "%s: can't reopen temporary file %s: ",
					progname, tfp->tf_filename);
			perror("");
			exit(1);
		}

#ifdef DBSORT
	dbset2up(tfp->tf_dbfd);
#endif

#ifdef DEBUG
	fprintf(stderr, "merging %s\n", tfp->tf_filename);
#endif

#ifdef SHUFFLE

	ntmp++;
	nte += tfp->tf_nents;

#else

#ifdef VMLIMIT
	/* XXX temporary -- until "aarrgh: case I can't handle" fixed */
	lmalreset();
#endif

#ifndef DBSORT
	tfp->tf_line = agetline2x(tfp->tf_fd, &r, (int *)NULL);
#else
	tfp->tf_dbnp = db_fgetent(tfp->tf_dbfd, dbd, (struct dbnode *)NULL, DB_CHECKRET, &r);
#endif

#ifndef DBSORT
	if(tfp->tf_line == NULL)
#else
	if(tfp->tf_dbnp == NULL)
#endif
		{
		switch(r)
			{
#ifndef DBSORT
			case AGL_EOF:
#else
			case DB_EOF:
#endif
#ifdef notdef
				/*
				 *  This *can* happen, in at least two cases:
				 *  1. we run out of memory in the
				 *     line-allocating step of input(),
				 *     that is, while reallocating the lines
				 *     array, before trying to read a line
				 *     (and if that line-reading attempt
				 *     ends up hitting EOF), or
				 *  2. we're using the #ifdef ESUITE/maxchunk
				 *     code, and ditto.
				 */

				fprintf(stderr,
					 "%s: oddity: temp file %s empty\n",
						progname, tfp->tf_filename);
#endif
#ifndef DBSORT
				(void)fclose(tfp->tf_fd);
#else
				(void)db_fclose(tfp->tf_dbfd);
#endif
#ifdef ESUITE
				ntmpfd--;
#endif
#ifdef DEBUG
				fprintf(stderr, "would remove %s\n",
							tfp->tf_filename);
#else
				if(remove(tfp->tf_filename) < 0)
					{
					fprintf(stderr,
					"%s: can't remove temporary file %s: ",
						progname, tfp->tf_filename);
					perror("");
					}
#endif
				free((char *)tfp);

				continue;

#ifndef DBSORT
			case AGL_NOMEM:
#else
			case DB_NOMEM:
#endif
				fprintf(stderr,
	 "%s: aarrgh: case I can't handle: out of memory during merge\n",
								progname);
				exit(1);

			default:
				fprintf(stderr,
					 "%s: error reading temp file %s\n",
						progname, tfp->tf_filename);
				exit(1);
			}
		}

	insert(tfp, &mergelist);
#endif
	}

#ifndef SHUFFLE

if(!usedmoretemps)
	domerge(mergelist, ofd, name);
else	{
	if(mergelist != NULL)
		{
		yetanothertemp(mergelist, &nxttmps);	/* merge last set */
		mergelist = NULL;
		}

	usetmp = nxttmps;
	nxttmps = NULL;
#ifdef DEBUG
	nstage++;
#endif
	goto again;
	}


#else

assert(!usedmoretemps);
domerge(usetmp, ofd, name);

#endif
}

domerge(mergelist, ofd, name)
struct tmpfile *mergelist;
FILE *ofd;
char *name;
{
#ifdef DBSORT
struct dbfd *dbofd;
#endif
struct tmpfile *tfp;
#ifdef SHUFFLE
struct tmpfile *nexttfp;
#endif
int r;

#ifdef DBSORT
dbofd = db_fdopen(ofd);
dbset2up(dbofd);
#endif

#ifndef SHUFFLE

/* now, repeatedly print first, re-reading from that file and re-inserting */

while(mergelist != NULL)
	{
	tfp = mergelist;
	mergelist = tfp->tf_next;

#ifndef DBSORT
	fprintf(ofd, "%s\n", tfp->tf_line);
#else
	db_putent(dbofd, tfp->tf_dbnp);
#endif

	if(ferror(ofd))		/* can't trust fprintf return value */
		{
		if(name == NULL)
			fprintf(stderr, "%s: output write error: ", progname);
		else	fprintf(stderr, "%s: error writing %s: ",
							progname, name);

		perror("");
		exit(1);
		}

#ifdef VMLIMIT
	/* XXX temporary -- until "aarrgh: case I can't handle" fixed */
	lmalreset();
#endif

#ifndef DBSORT
	free(tfp->tf_line);
	tfp->tf_line = agetline2x(tfp->tf_fd, &r, (int *)NULL);
#else
	db_nfree(tfp->tf_dbnp);
	tfp->tf_dbnp = db_fgetent(tfp->tf_dbfd, dbd, (struct dbnode *)NULL, DB_CHECKRET, &r);
#endif

#ifndef DBSORT
	if(tfp->tf_line == NULL)
#else
	if(tfp->tf_dbnp == NULL)
#endif
		{
		switch(r)
			{
#ifndef DBSORT
			case AGL_EOF:
#else
			case DB_EOF:
#endif
#ifndef DBSORT
				(void)fclose(tfp->tf_fd);
#else
				db_fclose(tfp->tf_dbfd);
#endif
#ifdef ESUITE
				ntmpfd--;
#endif
#ifdef DEBUG
				fprintf(stderr, "would remove %s\n",
							tfp->tf_filename);
#else
				if(remove(tfp->tf_filename) < 0)
					{
					fprintf(stderr,
					"%s: can't remove temporary file %s: ",
						progname, tfp->tf_filename);
					perror("");
					}
#endif
				free((char *)tfp);
				continue;

#ifndef DBSORT
			case AGL_NOMEM:
#else
			case DB_NOMEM:
#endif
				fprintf(stderr,
	 "%s: aarrgh: case I can't handle: out of memory during merge\n",
								progname);
				exit(1);

			default:
				fprintf(stderr,
					 "%s: error reading temp file %s\n",
						progname, tfp->tf_filename);
				exit(1);
			}
		}

	insert(tfp, &mergelist);
	}

#else /* above is !SHUFFLE, below is SHUFFLE */

while(nte > 0)
	{
#ifdef wrongway
	int i;
#else
	int x;
#endif

#ifdef DBSORT
	assert(0);	/* i.e. assert(!DBSORT) */
#endif

#ifdef wrongway
	r = randrange(ntmp);
	for(i = 0, tfp = tmpbase; i < r; i++, tfp = tfp->tf_next)
		;
#else
	r = randrange(nte);
	tfp = mergelist;
	x = tfp->tf_nents;
	while(tfp != NULL && (x <= r || !(tfp->tf_flags & TF_INUSE)))
		{
		tfp = tfp->tf_next;
		x += tfp->tf_nents;
		}
#endif

	assert(tfp != NULL);

#ifdef VMLIMIT
	/* XXX temporary -- until "aarrgh: case I can't handle" fixed */
	lmalreset();
#endif

	tfp->tf_line = agetline2x(tfp->tf_fd, &r, (int *)NULL);

	if(tfp->tf_line == NULL)
		{
		switch(r)
			{
			case AGL_EOF:
#ifndef wrongway
				assert(0);	/* i.e. "can't happen" */
#endif
/*				(void)fclose(tfp->tf_fd); */
				if(!(tfp->tf_flags & TF_EOF))	/* XXX awk */
					{
					if(remove(tfp->tf_filename) < 0)
						{
						fprintf(stderr,
					"%s: can't remove temporary file %s: ",
							progname,
							     tfp->tf_filename);
						perror("");
						}
/*					free((char *)tfp); */
#ifdef ESUITE
					ntmpfd--;
#endif
					}
#ifdef notdef
				if(!(tfp->tf_flags & TF_EOF))	/* XXX awkward */
					neof++;
#endif
				tfp->tf_flags |= TF_EOF;
				continue;

			case AGL_NOMEM:
				fprintf(stderr,
	 "%s: aarrgh: case I can't handle: out of memory during merge\n",
								progname);
				exit(1);

			default:
				fprintf(stderr,
					 "%s: error reading temp file %s\n",
						progname, tfp->tf_filename);
				exit(1);
			}
		}

	fprintf(ofd, "%s\n", tfp->tf_line);

	if(ferror(ofd))		/* can't trust fprintf return value */
		{
		if(name == NULL)
			fprintf(stderr, "%s: output write error: ", progname);
		else	fprintf(stderr, "%s: error writing %s: ",
							progname, name);

		perror("");
		exit(1);
		}

	free(tfp->tf_line);
	tfp->tf_nents--;
	nte--;
	}

for(tfp = mergelist; tfp != NULL; tfp = nexttfp)
	{
	nexttfp = tfp->tf_next;

	if(tfp->tf_flags & TF_INUSE)
		{
		(void)fclose(tfp->tf_fd);
#ifdef ESUITE
		ntmpfd--;
#endif
		if(remove(tfp->tf_filename) < 0)
			{
			fprintf(stderr, "%s: can't remove temporary file %s: ",
						progname, tfp->tf_filename);
			perror("");
			}
		}

	free(tfp);
	}

#endif /* SHUFFLE */
}

#ifndef SHUFFLE

static
insert(tfp, listp)
struct tmpfile *tfp;
struct tmpfile **listp;
{
struct tmpfile **tfpp;
int r;

for(tfpp = listp; *tfpp != NULL; tfpp = &(*tfpp)->tf_next)
	{
	/* worry about -u flag */

#ifndef DBSORT
	r = compare((char *)&tfp->tf_line, (char *)&(*tfpp)->tf_line);
#else
	r = dbcompare((char *)&tfp->tf_dbnp, (char *)&(*tfpp)->tf_dbnp);
#endif

	if(r <= 0)
		break;
	}

tfp->tf_next = *tfpp;
*tfpp = tfp;
}

#endif

#ifndef SHUFFLE	/* multi-stage not working for SHUFFLE yet */

/*
 *  Merge current opened temporary files, write to yet another temporary
 *  file, in the process closing the first set of temporary files,
 *  in preparation for another set.
 */

yetanothertemp(mergelist, nxttmpsp)
struct tmpfile *mergelist;
struct tmpfile **nxttmpsp;
{
struct tmpfile *nextemp;
FILE *fd;

fpunstash();

nextemp = tfalloc();

mktmpname(nextemp->tf_filename, 'a');

if((fd = fopen(nextemp->tf_filename, "w")) == NULL)
	{
	fprintf(stderr, "%s: can't open temporary file %s: ",
					progname, nextemp->tf_filename);
	perror("");
	exit(1);
	}

#ifdef ESUITE
ntmpfd++;
#endif

#ifdef DEBUG
fprintf(stderr, "using temporary file %s\n", nextemp->tf_filename);
#endif

nextemp->tf_flags |= TF_INUSE;

domerge(mergelist, fd, nextemp->tf_filename);

/* nextemp->tf_nents = nlines; */

if(fclose(fd) == EOF)
	{
	fprintf(stderr, "%s: error closing (writing?) temporary file %s: ",
					progname, nextemp->tf_filename);
	perror("");
	exit(1);
	}

#ifdef ESUITE
ntmpfd--;
#endif

nextemp->tf_next = *nxttmpsp;
*nxttmpsp = nextemp;

fpstash();
}

#endif

static FILE *stashfp;

fpstash()
{
/*
 *  Stash one away, because by the time we need an intermediate
 *  output file to do an n-way merge, we'll just have run out of 'em
 *  by opening a bunch for input...
 */

stashfp = fopen("/dev/null", "r");
}

fpunstash()
{
fclose(stashfp);
}
